# This Python script processes and analyzes Origin-Destination (OD) data to calculate 
# and visualize the coefficient of variation (CV) for motor vehicle production and 
# attraction across different spatial levels. The results are saved as CSV files and 
# visualized using violin plots to compare variability across zones.

# Import necessary libraries for data processing, numerical operations, and visualization
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle
import time

# Record the start time of the script execution
start_time = time.time()

# _______________ Loading Data ______________________
# Load OD pair production data from a pickle file
fileNAME = 'production_attraction'
with open(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/ODpairDemand{fileNAME}.pickle', 'rb') as handle:
    ODpairProduction = pickle.load(handle)

# Load OD pair attraction data from a pickle file
fileNAME = 'attraction'
with open(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/ODpairDemand{fileNAME}.pickle', 'rb') as handle:
    ODpairAttraction = pickle.load(handle)

# Load the list of OD zones from a CSV file
ODzones = pd.read_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/mezuro_areas_2017.CSV", usecols=["mzr_id"])["mzr_id"].tolist()

# ___________________ Pre-processing: Generating Detailed Motor-Vehicle Production-Attraction Dictionary ___________________

# Create a combined dictionary with production and attraction data for each zone
ODpairDemand = {}
for i in ODpairProduction:
    # Combine production and attraction data for each date
    ODpairDemand[i] = pd.DataFrame(ODpairProduction[i]['Aantal reizen motorvoertuigen 12+']).rename(
        columns={'Aantal reizen motorvoertuigen 12+': "production"}).sort_index()
    ODpairDemand[i]["attraction"] = pd.DataFrame(ODpairAttraction[i]['Aantal reizen motorvoertuigen 12+']).sort_index()

# Save the combined production-attraction dictionary to a pickle file
fileNAME = 'ProductionAttraction_motorVehicle'
with open(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/{fileNAME}.pickle', 'wb') as handle:
    pickle.dump(ODpairDemand, handle, protocol=pickle.HIGHEST_PROTOCOL)

# Clean up memory by deleting the original production and attraction data
del ODpairProduction, ODpairAttraction

# _____________________ Pre-processing: Generating Daily Motor-Vehicle Production-Attraction Dictionary for Each Zone ______________

def ODdemandGenerator_daily(ODpairDemand, ODzones):
    """
    Generates a dictionary of daily production and attraction data for each zone.

    :param ODpairDemand: Dictionary containing combined OD pair demand data.
    :param ODzones: List of OD zones.
    :return: Dictionary with daily production and attraction for each zone.
    """
    kk = dict.fromkeys(ODpairDemand)
    for j in ODpairDemand.keys():
        kk[j] = {}
        for i in ODzones:
            kk[j][i] = {"production": sum(ODpairDemand[j].loc[i]["production"]),
                        "attraction": sum(ODpairDemand[j].loc[i]["attraction"])}
    return kk

# Generate the daily production-attraction DataFrame
ODdemand_daily_df = pd.DataFrame.from_dict(ODdemandGenerator_daily(ODpairDemand=ODpairDemand, ODzones=ODzones))

# Clean up memory by deleting the combined ODpairDemand dictionary
del ODpairDemand

# _____________________ Generating Daily Motor-Vehicle Production-Attraction Coefficient of Variation (CV) for Each Zone ______________

# Initialize a DataFrame to store the CV values for production and attraction
CVdaily = pd.DataFrame(columns=["production", "attraction"], index=ODzones)

# Calculate the CV for each zone based on daily production and attraction data
for i in ODdemand_daily_df.index:
    kk = ODdemand_daily_df.loc[i, :]
    prod = []
    att = []
    for j in kk.keys():  # Iterate over each date
        prod.append(kk[j]["production"])
        att.append(kk[j]["attraction"])
    CVdaily.loc[i, "production"] = np.std(prod) / np.mean(prod)
    CVdaily.loc[i, "attraction"] = np.std(att) / np.mean(att)

# Reset the index and rename the columns for better readability
CVdaily = CVdaily.reset_index()
CVdaily.columns = ['mzr_id', 'production', 'attraction']

# Save the CV data to a CSV file
CVdaily.to_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/CVdaily_mzr.CSV", header=True, index=False)
CVdaily["spatialLevel"] = "level5"

# _______________________________ Split Production and Attraction Data _________________________________

# Separate the daily production and attraction data into two DataFrames
ODproduction_daily_df = pd.DataFrame(columns=ODdemand_daily_df.columns, index=ODdemand_daily_df.index)
ODattraction_daily_df = pd.DataFrame(columns=ODdemand_daily_df.columns, index=ODdemand_daily_df.index)

# Populate the separate DataFrames with the respective production and attraction data
for i in ODdemand_daily_df.index:
    for j in ODdemand_daily_df.columns:
        ODproduction_daily_df.loc[i, j] = ODdemand_daily_df.loc[i, j]["production"]
        ODattraction_daily_df.loc[i, j] = ODdemand_daily_df.loc[i, j]["attraction"]

# Clean up memory by deleting the combined daily demand DataFrame
del ODdemand_daily_df

# _________________________________ Spatial Aggregation ______________________________________

# Load zoning information to aggregate data at different spatial levels
zoning = pd.read_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/mezuro_areas_2017.CSV",
                     usecols=['mzr_id', 'prov_id', 'wnp_id', 'gem_id'], index_col="mzr_id").sort_index()

def CVcalculator(production, attraction):
    """
    Calculates the Coefficient of Variation (CV) for production and attraction data.

    :param production: DataFrame containing production data.
    :param attraction: DataFrame containing attraction data.
    :return: DataFrame with CV values for production and attraction.
    """
    kkk = pd.DataFrame(columns=["production", "attraction"], index=production.index)
    for i in production.index:  # Iterate over each zone
        kkp = production.loc[i, :]
        kka = attraction.loc[i, :]
        prod = []
        att = []
        for j in kkp.index:  # Iterate over each date
            prod.append(kkp[j])
            att.append(kka[j])
        kkk.loc[i, "production"] = np.std(prod) / np.mean(prod)
        kkk.loc[i, "attraction"] = np.std(att) / np.mean(att)
    return kkk

# Aggregate production and attraction data at different spatial levels (wnp_id, gem_id, prov_id)
kp = ODproduction_daily_df.join(zoning['wnp_id']).sort_index()
ka = ODattraction_daily_df.join(zoning['wnp_id']).sort_index()
ODproduction_daily_wnp = kp.groupby(by=['wnp_id'], sort=True).sum()
ODattraction_daily_wnp = ka.groupby(by=['wnp_id'], sort=True).sum()
CVdaily_wnp = CVcalculator(production=ODproduction_daily_wnp, attraction=ODattraction_daily_wnp)
CVdaily_wnp = CVdaily_wnp.reset_index()
CVdaily_wnp.to_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/CVdaily_wnp.CSV", header=True, index=False)
CVdaily_wnp["spatialLevel"] = "level4"

kp = ODproduction_daily_df.join(zoning['gem_id']).sort_index()
ka = ODattraction_daily_df.join(zoning['gem_id']).sort_index()
ODproduction_daily_gem = kp.groupby(by=['gem_id'], sort=True).sum()
ODattraction_daily_gem = ka.groupby(by=['gem_id'], sort=True).sum()
CVdaily_gem = CVcalculator(production=ODproduction_daily_gem, attraction=ODattraction_daily_gem)
CVdaily_gem = CVdaily_gem.reset_index()
CVdaily_gem.to_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/CVdaily_gem.CSV", header=True, index=False)
CVdaily_gem["spatialLevel"] = "level3"

kp = ODproduction_daily_df.join(zoning['prov_id']).sort_index()
ka = ODattraction_daily_df.join(zoning['prov_id']).sort_index()
ODproduction_daily_prov = kp.groupby(by=['prov_id'], sort=True).sum()
ODattraction_daily_prov = ka.groupby(by=['prov_id'], sort=True).sum()
CVdaily_prov = CVcalculator(production=ODproduction_daily_prov, attraction=ODattraction_daily_prov)
CVdaily_prov = CVdaily_prov.reset_index()
CVdaily_prov.to_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/CVdaily_prov.CSV", header=True, index=False)
CVdaily_prov["spatialLevel"] = "level2"

# Calculate CV for the entire dataset
ODproduction_daily_entire = ODproduction_daily_df.sum(axis=0)
ODattraction_daily_entire = ODattraction_daily_df.sum(axis=0)
CVdaily_entire = pd.DataFrame({"dissolve": [1], "production": [np.std(ODproduction_daily_entire) / np.mean(
    ODproduction_daily_entire)], "attraction": [np.std(ODattraction_daily_entire) / np.mean(ODattraction_daily_entire)]})
CVdaily_entire.to_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/CVdaily_entire.CSV", header=True, index=False)
CVdaily_entire["spatialLevel"] = "level1"

# Combine CV data across different spatial levels
kk = CVdaily_entire.append(CVdaily_prov).append(CVdaily_gem).append(CVdaily_wnp).append(CVdaily)
kk = kk.loc[:, ["production", "attraction", "spatialLevel"]]

# Prepare data for visualization
kkp = kk.loc[:, ["production", "spatialLevel"]]
kkp["demand type"] = "production"
kkp.columns = ['CV', 'spatialLevel', 'demand type']
kka = kk.loc[:, ["attraction", "spatialLevel"]]
kka["demand type"] = "attraction"
kka.columns = ['CV', 'spatialLevel', 'demand type']
kk = kkp.append(kka)
kk["CV"] = kk["CV"].astype('float64')

# _______________________________ Visualization ______________________________________
sns.set_theme(style="darkgrid")
sns.set(font_scale=1.8)
fig, axs = plt.subplots(figsize=(9, 7))
axs.set_ylim(0.06, 0.22)

# Create a violin plot to visualize the CV distribution by spatial level
sns.violinplot(x="spatialLevel", y="CV", hue="demand type", data=kk,
               palette="Reds", split=True, inner="quartile", bw=.2)
axs.legend(title="")
plt.show()

# Save the plot as a PNG file
fileNAME = "spatialAggBoxplot"
plt.savefig(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/{fileNAME}.png', dpi=150)
