# This Python script analyzes the temporal variability of motor vehicle production 
# and attraction across different zones by calculating the coefficient of variation 
# (CV) for various time aggregation intervals. The results are visualized using violin 
# plots and saved as CSV and PNG files for further analysis.

import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
import pickle
import time

# Record the start time of the script execution
start_time = time.time()

# Load the OD pair production data from a pickle file
fileNAME = 'production_attraction'
with open(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/ODpairDemand{fileNAME}.pickle', 'rb') as handle:
    ODpairProduction = pickle.load(handle)  # This is production data

# Load the OD pair attraction data from a pickle file
fileNAME = 'attraction'
with open(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/ODpairDemand{fileNAME}.pickle', 'rb') as handle:
    ODpairAttraction = pickle.load(handle)  # This is attraction data

# Load the list of OD zones and sort them
ODzones = pd.read_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/mezuro_areas_2017.CSV", usecols=["mzr_id"])["mzr_id"].tolist()
ODzones.sort()

# ____________________________________________________________________________
def tempAgg(ODpairProduction, ODpairAttraction, ODzones, aggInterval):
    """
    Aggregates OD pair production and attraction data over specified time intervals and calculates the CV.

    :param ODpairProduction: Dictionary containing OD pair production data.
    :param ODpairAttraction: Dictionary containing OD pair attraction data.
    :param ODzones: List of OD zones.
    :param aggInterval: Aggregation interval in hours.
    :return: DataFrame containing CV values for production and attraction.
    """
    CVtableP = pd.DataFrame(index=ODzones, columns=["CV"])
    CVtableA = pd.DataFrame(index=ODzones, columns=["CV"])

    # Calculate the CV for each zone based on the specified aggregation interval
    for j in ODzones:
        kkp = np.array([])
        kka = np.array([])
        for i in ODpairProduction.keys():
            kkkp = ODpairProduction[i].loc[j].loc[:, "Aantal reizen motorvoertuigen 12+"]
            kkp = np.hstack((kkkp, np.add.reduceat(np.asarray(kkkp), np.arange(0, len(kkkp), aggInterval))))
            kkka = ODpairAttraction[i].loc[j].loc[:, "Aantal reizen motorvoertuigen 12+"]
            kka = np.hstack((kkka, np.add.reduceat(np.asarray(kkka), np.arange(0, len(kkka), aggInterval))))
        CVtableP.loc[j, "CV"] = kkp.std() / kkp.mean()
        CVtableA.loc[j, "CV"] = kka.std() / kka.mean()

    CVtableP["demand type"] = "production"
    CVtableA["demand type"] = "attraction"
    CVtable = CVtableP.append(CVtableA)
    CVtable["aggregation interval [hr]"] = f"{aggInterval}"
    return CVtable

# Calculate CVs for various time aggregation intervals
CV_1h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=1)
CV_2h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=2)
CV_3h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=3)
CV_4h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=4)
CV_5h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=5)
CV_6h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=6)
CV_12h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=12)
CV_24h = tempAgg(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=24)

# Combine the CV data for all aggregation intervals
metaCV = CV_1h.append(CV_2h).append(CV_3h).append(CV_4h).append(CV_5h).append(CV_6h).append(CV_12h).append(CV_24h)
metaCV["CV"] = metaCV["CV"].astype("float64")

# Plot the results using a violin plot to visualize the distribution of CVs across aggregation intervals
sns.set_theme(style="darkgrid")
sns.set(font_scale=1.8)
fig, axs = plt.subplots(figsize=(9, 7))
sns.violinplot(x="aggregation interval [hr]", y="CV", hue="demand type", data=metaCV,
               palette="Reds", split=True, inner="quartile", bw=.2)
plt.show()

# Save the plot as a PNG file
fileNAME = "temporalAggBoxplot_entire"
plt.savefig(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/{fileNAME}.png', dpi=150)

# __________________________________________________________________________________

def tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones, aggInterval):
    """
    Aggregates OD pair production and attraction data over peak hours and calculates the CV.

    :param ODpairProduction: Dictionary containing OD pair production data.
    :param ODpairAttraction: Dictionary containing OD pair attraction data.
    :param ODzones: List of OD zones.
    :param aggInterval: List of peak hours to aggregate over.
    :return: DataFrame containing CV values for production and attraction during peak hours.
    """
    CVtableP = pd.DataFrame(index=ODzones, columns=["CV"])
    CVtableA = pd.DataFrame(index=ODzones, columns=["CV"])

    # Calculate the CV for each zone based on the specified peak hours
    for j in ODzones:
        kkp = np.array([])
        kka = np.array([])
        for i in ODpairProduction.keys():
            kkp = np.hstack((kkp, ODpairProduction[i].loc[j].loc[aggInterval, "Aantal reizen motorvoertuigen 12+"].sum()))
            kka = np.hstack((kka, ODpairAttraction[i].loc[j].loc[aggInterval, "Aantal reizen motorvoertuigen 12+"].sum()))
        CVtableP.loc[j, "CV"] = kkp.std() / kkp.mean()
        CVtableA.loc[j, "CV"] = kka.std() / kka.mean()

    CVtableP["demand type"] = "production"
    CVtableA["demand type"] = "attraction"
    CVtable = CVtableP.append(CVtableA)
    CVtable["aggregation interval"] = f"{aggInterval[0]}-{aggInterval[-1] + 1}"
    return CVtable

def tempAgg_entire(ODpairProduction, ODpairAttraction, ODzones):
    """
    Aggregates OD pair production and attraction data over the entire day and calculates the CV.

    :param ODpairProduction: Dictionary containing OD pair production data.
    :param ODpairAttraction: Dictionary containing OD pair attraction data.
    :param ODzones: List of OD zones.
    :return: DataFrame containing CV values for production and attraction for the entire day.
    """
    CVtableP = pd.DataFrame(index=ODzones, columns=["CV"])
    CVtableA = pd.DataFrame(index=ODzones, columns=["CV"])

    # Calculate the CV for each zone based on the entire day's data
    for j in ODzones:
        kkp = np.array([])
        kka = np.array([])
        for i in ODpairProduction.keys():
            kkp = np.hstack((kkp, ODpairProduction[i].loc[j].loc[:, "Aantal reizen motorvoertuigen 12+"].sum()))
            kka = np.hstack((kka, ODpairAttraction[i].loc[j].loc[:, "Aantal reizen motorvoertuigen 12+"].sum()))
        CVtableP.loc[j, "CV"] = kkp.std() / kkp.mean()
        CVtableA.loc[j, "CV"] = kka.std() / kka.mean()

    CVtableP["demand type"] = "production"
    CVtableA["demand type"] = "attraction"
    CVtable = CVtableP.append(CVtableA)
    CVtable["aggregation interval"] = "entire day"
    return CVtable

# Calculate CVs for various peak hour aggregations
CV_M7 = tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=[7])
CV_M7_8 = tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=[7, 8])
CV_M7_9 = tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=[7, 8, 9])
CV_M7_10 = tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=[7, 8, 9, 10])
CV_M7_11 = tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones=ODzones, aggInterval=[7, 8, 9, 10, 11])
CV_M7_19 = tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones=ODzones,
                        aggInterval=[7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18])
CV_M6_23 = tempAgg_peak(ODpairProduction, ODpairAttraction, ODzones=ODzones,
                        aggInterval=[6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23])
CV_entire = tempAgg_entire(ODpairProduction, ODpairAttraction, ODzones=ODzones)

# Save the CV data for the entire day
fileNAME = "dailyCVtable_entireNL"
with open(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/{fileNAME}.pickle', 'wb') as handle:
    pickle.dump(CV_entire, handle, protocol=pickle.HIGHEST_PROTOCOL)

# Combine all CV data
metaCV = CV_entire.append(CV_M6_23).append(CV_M7_19).append(CV_M7_11).append(CV_M7_10).append(CV_M7_9).append(
    CV_M7_8).append(CV_M7)
metaCV["CV"] = metaCV["CV"].astype("float64")

# Plot the results using a violin plot to visualize the distribution of CVs across aggregation intervals
sns.set_theme(style="darkgrid")
sns.set(font_scale=1.7)
fig, axs = plt.subplots(figsize=(12, 7))
axs.set_ylim(0.03, 0.53)
sns.violinplot(x="aggregation interval", y="CV", hue="demand type", data=metaCV,
               palette="Reds", split=True, inner="quartile", bw=.2)
axs.legend(title="")
plt.show()

# Save the plot as a PNG file
fileNAME = "temporalAggBoxplot_MP"
plt.savefig(f'Y:/ZahraEftekhar/dittlabData/demandAnalysis/{fileNAME}.png', dpi=150)
