# This Python script processes Origin-Destination (OD) data to generate matrices and 
# statistical analyses, including daily and hourly production metrics for different zones. 
# It uses various functions to aggregate and analyze transportation demand data, outputting 
# the results as CSV files for further analysis.

# Import necessary libraries for data processing, numerical operations, and visualization
import json
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from os import listdir, path  # For interacting with the file system
import pickle  # For serializing and deserializing Python objects
import concurrent.futures  # For running tasks concurrently
from datetime import datetime  # For handling date and time operations
import time  # For tracking time

# Set the filename for output or processing and start the timer
fileNAME = 'production_attraction'
start_time = time.time()

def ODmatrixGenerator(fileAddress, zoningCSVaddress, zoningColName):
    """
    This function generates an OD matrix from a given OD data file and a zoning file.
    :param fileAddress: the address of the OD data file
    :param zoningCSVaddress: the address of the zoning file
    :param zoningColName: the name of the column in the zoning file that contains the zone names
    :return: an OD matrix
    
    """
    # Read the OD data file
    ODdata = pd.read_csv(fileAddress, sep=";")
    
    # Read the zoning file and extract the zone names
    tazNames = pd.read_csv(zoningCSVaddress, usecols=[zoningColName])
    
    # Create a series with a zero entry to represent the base zone
    zoneZero = pd.Series(0)
    
    # Combine the base zone and the zone names into one series
    matrixRowColNames = tuple(zoneZero.append(tazNames))
    
    # Determine the size of the matrix (number of zones)
    odsize = len(matrixRowColNames)
    
    # Initialize an empty OD matrix with zeros
    ODMatrix_empty = pd.DataFrame(np.zeros((odsize, odsize), dtype=np.float), columns=matrixRowColNames,
                               index=matrixRowColNames)

def addressGenerator(mainPath, fileName,ODzones):
    """
    This function generates the address of the OD data file and the OD zones
    :param mainPath: the main path of the OD data files
    :param fileName: the name of the OD data file
    :param ODzones: the list of OD zones
    :return: a dictionary including the address of the OD data file and the list of OD zones
    
    """
    # Check if the file exists at the given path
    if path.isfile(path.join(mainPath, fileName)):
        # Return the full file path and OD zones in a dictionary
        fileAddress = path.join(mainPath, fileName)
        return {"address": fileAddress, "ODzones": ODzones}

# Define paths where OD data might be located
paths = ['D:/MiRRORS data/MiRRORS data', 'C:/Users/zahraeftekhar/Documents/MiRRORS data']
addresses = {}  # Dictionary to hold addresses of OD data files

# Load OD zones and population data from a CSV file
ODzones = (pd.read_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/mezuro_areas_2017.CSV",
                      usecols=["mzr_id"]))["mzr_id"].tolist()

population = (pd.read_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/mezuro_areas_2017.CSV",
                      usecols=["mzr_id","inhabitant"])).sort_values(by=["mzr_id"])
population = population.set_index(["mzr_id"], drop=True)  # Set mzr_id as the index for the population DataFrame

# Generate addresses for OD data files
j = 0
for i in paths:
    for entry in listdir(i):
        addresses[j] = addressGenerator(mainPath=i, fileName=entry, ODzones=ODzones)
        j += 1


def ODpairDemandGenerator(items):
    """
    This function generates the OD pair demand for a given OD data file and OD zones
    :param items: a dictionary including the address of the OD data file and the list of OD zones
    :return: an OD pair demand dataframe
    
    """
    # Read the OD data file
    ODdata = pd.read_csv(items["address"], sep=";")
    
    # Group the data by origin and hour, aggregating travel counts
    kk = ((ODdata.groupby(by=['Herkomst id', 'Uur'], sort=True)).agg({
        'Aantal reizen motorvoertuigen 12+': 'sum',
        'Aantal reizen trein 12+': 'sum',
        'Aantal motorvoertuigen 12+': 'sum'}).reset_index())
    
    # Convert the aggregated data to a NumPy array
    kkkk = (kk.loc[:, ['Aantal reizen motorvoertuigen 12+',
                       'Aantal reizen trein 12+', 'Aantal motorvoertuigen 12+']]).to_numpy()
    
    # Create a multi-index for the DataFrame using origin and hour
    arrays = [kk["Herkomst id"].tolist(), kk["Uur"].tolist()]
    index = pd.MultiIndex.from_arrays(arrays, names=["origin", "Uur"])
    
    # Create the final OD pair demand DataFrame
    kkk = pd.DataFrame(kkkk, index=index, copy=True, columns=['Aantal reizen motorvoertuigen 12+',
                                                              'Aantal reizen trein 12+', 'Aantal motorvoertuigen 12+'])
    return kkk

# Generate a date range for March 2017
dateRange = pd.date_range(start='03/01/2017', end='03/31/2017')

# Load precomputed OD pair demand data from a pickle file
with open('Y:/ZahraEftekhar/dittlabData//demandAnalysis/ODpairDemand{a}.pickle'.format(a=fileNAME), 'rb') as handle:
    ODpairDemand = pickle.load(handle)

# Extract OD data for a specific date
ODdata = ODpairDemand[pd.to_datetime('2017-03-01')]

def ODproductionGenerator_daily(ODpairDemand,ODzones,population):

    """
    Generates daily production data for motor vehicles and trains, including mean and standard deviation.

    :param ODpairDemand: Dictionary containing OD pair demand data.
    :param ODzones: List of OD zones.
    :param population: DataFrame containing population data.
    :return: Dictionary with daily production statistics for each zone.
    """
    # Initialize a dictionary to store production data for each zone
    kk = dict.fromkeys(ODzones)
    
    # Loop over each day in the ODpairDemand keys
    for j in ODpairDemand.keys():
        ODdata = ODpairDemand[j]
        
        # Loop over each zone in ODzones
        for i in ODzones:
            # Calculate mean, standard deviation, and ratio of sd to mean for each zone
            kk[i] = {j: {"means": ODdata.loc[i].mean(), "sd": ODdata.loc[i].std(),
                         "sd/mean": ODdata.loc[i].std() / ODdata.loc[i].mean(),
                         "population": population.loc[i, "inhabitant"]}}
    return kk

# Generate daily OD production data
ODproduction_daily = ODproductionGenerator_daily(ODpairDemand=ODpairDemand, ODzones=ODzones, population=population)


def ODproductionGenerator_entire_hourly(ODpairDemand,ODzones): 

    """ 
    This function generates the mean and standard deviation of the production (motor-vehicle) for each day and each zone
    :param ODpairDemand: the OD pair demand dataframe
    :param ODzones: the list of OD zones
    :return: a dictionary including the mean and standard deviation of the production (motor-vehicle) for each day and each zone
    
    """
    # Initialize dictionaries to store concatenated hourly data and the final statistics
    kk = dict.fromkeys(ODzones)
    kkk = dict.fromkeys(ODzones)
    
    # Loop over each zone to concatenate data across all days
    for i in ODzones:
        kk[i] = np.array([])
        for j in ODpairDemand.keys():
            ODdata = ODpairDemand[j]
            kk[i] = np.concatenate((kk[i], np.asarray(ODdata.loc[i, "Aantal reizen motorvoertuigen 12+"])))
    
    # Calculate the standard deviation to mean ratio for each zone
    for i in ODzones:
        kkk[i] = [kk[i].std() / kk[i].mean()]

    return kkk

# Generate hourly production statistics across all days
ODproduction_entire_hourly = ODproductionGenerator_entire_hourly(ODpairDemand=ODpairDemand, ODzones=ODzones)
ODproduction_entire_hourly_df = pd.DataFrame.from_dict(ODproduction_entire_hourly)
ODproduction_entire_hourly_df = pd.to_numeric(ODproduction_entire_hourly_df.iloc[0, :])


def ODproductionGenerator_weekday_AveragePeakHour7_10(ODpairDemand,ODzones): 
    """
    Generate OD production for weekdays during average peak hours (7-10 AM).

    Parameters:
    ODpairDemand (dict): Dictionary containing OD pair demand data.
    ODzones (list): List of OD zones.

    Returns:
    dict: Dictionary containing OD production data for the specified time period.
    """
    # Initialize a dictionary to store data for each zone
    kk = dict.fromkeys(ODzones)
    
    # Loop over each zone to accumulate peak hour data on weekdays
    for i in ODzones:
        kk[i] = []
        for j in ODpairDemand.keys():
            if j.weekday() < 5:  # Check if the day is a weekday
                ODdata = ODpairDemand[j]
                # Append mean values of peak hours (7-10 AM)
                kk[i] = kk[i] + [(np.asarray(ODdata.loc[(i, (7, 8, 9)), "Aantal reizen motorvoertuigen 12+"])).mean()]
    
    # Calculate the standard deviation to mean ratio for each zone
    kkk = dict.fromkeys(ODzones)
    for i in ODzones:
        kkk[i] = [np.std(kk[i]) / np.mean(kk[i])]

    return kkk

# Generate weekday peak hour production statistics
ODproduction_weekday_AveragePeakhour7_10 = ODproductionGenerator_weekday_AveragePeakHour7_10(ODpairDemand=ODpairDemand, ODzones=ODzones)

# Convert the resulting dictionary to a DataFrame and sort by index
ODproduction_weekday_AveragePeakhour7_10 = pd.DataFrame.from_dict(ODproduction_weekday_AveragePeakhour7_10, columns="sd2mean")
ODproduction_weekday_AveragePeakhour7_10_sorted = (ODproduction_weekday_AveragePeakhour7_10.sort_index(axis=1)).transpose()

# Add the zone ID as a column in the sorted DataFrame
ODproduction_weekday_AveragePeakhour7_10_sorted["mzr_id"] = ODproduction_weekday_AveragePeakhour7_10_sorted.index

# Save the sorted DataFrame to a CSV file
ODproduction_weekday_AveragePeakhour7_10_sorted.to_csv(
    "Y:/ZahraEftekhar/dittlabData/demandAnalysis/sd2meanPeakAverage7_10.CSV", index=False, header=True)