# This Python script processes transportation Origin-Destination (OD) data to generate 
# matrices and analyze demand, focusing on calculating the average production and 
# attraction during peak hours for different zones. The results are saved as CSV files, 
# which can be further analyzed or visualized.


# Import necessary libraries for data processing, numerical operations, and visualization
import json
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from os import listdir, path  # For interacting with the file system
import pickle  # For serializing and deserializing Python objects
import concurrent.futures  # For running tasks concurrently
from datetime import datetime  # For handling date and time operations
import time  # For tracking time

# Set the filename for output or processing and start the timer
fileNAME = 'attraction'
start_time = time.time()

def ODmatrixGenerator(fileAddress, zoningCSVaddress, zoningColName):
    """
    Generates an OD matrix from a given OD data file and a zoning file.

    :param fileAddress: Path to the OD data file.
    :param zoningCSVaddress: Path to the zoning file.
    :param zoningColName: Name of the column in the zoning file containing the zone names.
    :return: An OD matrix as a Pandas DataFrame.
    """
    # Read the OD data file
    ODdata = pd.read_csv(fileAddress, sep=";")
    
    # Read the zoning file and extract the zone names
    tazNames = pd.read_csv(zoningCSVaddress, usecols=[zoningColName])
    
    # Create a series with a zero entry to represent the base zone
    zoneZero = pd.Series(0)
    
    # Combine the base zone and the zone names into one series
    matrixRowColNames = tuple(zoneZero.append(tazNames))
    
    # Determine the size of the matrix (number of zones)
    odsize = len(matrixRowColNames)
    
    # Initialize an empty OD matrix with zeros
    ODMatrix_empty = pd.DataFrame(np.zeros((odsize, odsize), dtype=np.float), columns=matrixRowColNames,
                               index=matrixRowColNames)

def addressGenerator(mainPath, fileName, ODzones):
    """
    Generates the file path and OD zones for a given OD data file.

    :param mainPath: Directory path containing the OD data files.
    :param fileName: Name of the OD data file.
    :param ODzones: List of OD zones.
    :return: A dictionary with the file path and OD zones.
    """
    # Check if the file exists at the given path
    if path.isfile(path.join(mainPath, fileName)):
        # Return the full file path and OD zones in a dictionary
        fileAddress = path.join(mainPath, fileName)
        return {"address": fileAddress, "ODzones": ODzones}

# Define paths where OD data might be located
paths = ['D:/MiRRORS data/MiRRORS data', 'C:/Users/zahraeftekhar/Documents/MiRRORS data']
addresses = {}  # Dictionary to hold addresses of OD data files

# Load OD zones and population data from a CSV file
ODzones = (pd.read_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/mezuro_areas_2017.CSV",
                      usecols=["mzr_id"]))["mzr_id"].tolist()

population = (pd.read_csv("Y:/ZahraEftekhar/dittlabData/demandAnalysis/mezuro_areas_2017.CSV",
                      usecols=["mzr_id","inhabitant"])).sort_values(by=["mzr_id"])
population = population.set_index(["mzr_id"], drop=True)  # Set mzr_id as the index for the population DataFrame

# Generate addresses for OD data files
j = 0
for i in paths:
    for entry in listdir(i):
        addresses[j] = addressGenerator(mainPath=i, fileName=entry, ODzones=ODzones)
        j += 1

def ODpairDemandGenerator(items):
    """
    Generates an OD pair demand DataFrame from the OD data file.

    :param items: Dictionary containing the file path and OD zones.
    :return: A DataFrame with OD pair demand data, grouped by destination and hour.
    """
    # Read the OD data file
    ODdata = pd.read_csv(items["address"], sep=";")
    
    # Group the data by destination and hour, aggregating travel counts
    kk = ((ODdata.groupby(by=['Bestemming id', 'Uur'], sort=True)).agg({
        'Aantal reizen motorvoertuigen 12+': 'sum',
        'Aantal reizen trein 12+': 'sum',
        'Aantal motorvoertuigen 12+': 'sum'}).reset_index())
    
    # Convert the aggregated data to a NumPy array
    kkkk = (kk.loc[:, ['Aantal reizen motorvoertuigen 12+',
                       'Aantal reizen trein 12+', 'Aantal motorvoertuigen 12+']]).to_numpy()
    
    # Create a multi-index for the DataFrame using destination and hour
    arrays = [kk["Bestemming id"].tolist(), kk["Uur"].tolist()]
    index = pd.MultiIndex.from_arrays(arrays, names=["destination", "Uur"])
    
    # Create the final OD pair demand DataFrame
    kkk = pd.DataFrame(kkkk, index=index, copy=True, columns=['Aantal reizen motorvoertuigen 12+',
                                                              'Aantal reizen trein 12+', 'Aantal motorvoertuigen 12+'])
    return kkk

# Generate a date range for March 2017
dateRange = pd.date_range(start='03/01/2017', end='03/31/2017')

# Load precomputed OD pair demand data from a pickle file
with open('Y:/ZahraEftekhar/dittlabData//demandAnalysis/ODpairDemand{a}.pickle'.format(a=fileNAME), 'rb') as handle:
    ODpairDemand = pickle.load(handle)

def ODattractionGenerator_weekday_AveragePeakHour7_10(ODpairDemand, ODzones):
    """
    Generates production statistics for weekdays during average peak hours (7-10 AM).

    :param ODpairDemand: Dictionary containing OD pair demand data.
    :param ODzones: List of OD zones.
    :return: Dictionary with production statistics for peak hours on weekdays.
    """
    # Initialize a dictionary to store data for each zone
    kk = dict.fromkeys(ODzones)
    
    # Loop over each zone to accumulate peak hour data on weekdays
    for i in ODzones:
        kk[i] = []
        for j in ODpairDemand.keys():
            if j.weekday() < 5:  # Check if the day is a weekday
                ODdata = ODpairDemand[j]
                # Append mean values of peak hours (7-10 AM)
                kk[i] = kk[i] + [(np.asarray(ODdata.loc[(i, (7, 8, 9)), "Aantal reizen motorvoertuigen 12+"])).mean()]
    
    # Calculate the standard deviation to mean ratio for each zone
    kkk = dict.fromkeys(ODzones)
    for i in ODzones:
        kkk[i] = [np.std(kk[i]) / np.mean(kk[i])]

    return kkk

# Generate weekday peak hour production statistics
ODproduction_weekday_AveragePeakhour7_10 = ODattractionGenerator_weekday_AveragePeakHour7_10(ODpairDemand=ODpairDemand, ODzones=ODzones)

# Convert the resulting dictionary to a DataFrame and sort by index
ODproduction_weekday_AveragePeakhour7_10 = pd.DataFrame.from_dict(ODproduction_weekday_AveragePeakhour7_10)

ODproduction_weekday_AveragePeakhour7_10_sorted = (ODproduction_weekday_AveragePeakhour7_10.sort_index(axis=1)).transpose()
ODproduction_weekday_AveragePeakhour7_10_sorted.columns = ['sd2mean']
ODproduction_weekday_AveragePeakhour7_10_sorted["mzr_id"] = ODproduction_weekday_AveragePeakhour7_10_sorted.index

# Save the sorted DataFrame to a CSV file
ODproduction_weekday_AveragePeakhour7_10_sorted.to_csv(
    "Y:/ZahraEftekhar/dittlabData/demandAnalysis/sd2meanPeakAverage_attraction_7_10.CSV", index=False, header=True)

# Import additional libraries for potential further analysis (e.g., Bayesian modeling)
import matplotlib as plt
import pgmpy.models as BayesianModel
from pgmpy.inference import VariableElimination
